package de.ozgcloud.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.common.attached_item.AttachedItemTestFactory;
import de.ozgcloud.document.bescheid.Bescheid;
import de.ozgcloud.document.bescheid.BescheidService;
import de.ozgcloud.document.bescheid.vorgang.VorgangId;
import de.ozgcloud.document.common.attached_item.AttachedItemMapper;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import io.grpc.stub.StreamObserver;

class BescheidGrpcServiceTest {

	@Spy
	@InjectMocks
	private BescheidGrpcService grpcService;

	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private AttachedItemMapper attachedItemMapper;
	@Mock
	private BescheidService service;
	@Mock
	private BescheidMapper mapper;

	@DisplayName("Get bescheid draft")
	@Nested
	class TestGetBescheidDraft {

		@Mock
		private StreamObserver<GrpcGetBescheidDraftResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetBescheidDraftResponse> responseCaptor;

		private final GrpcGetBescheidDraftRequest request = GrpcGetBescheidDraftRequestTestFactory.create();

		private final GrpcGetBescheidDraftResponse response = GrpcGetBescheidDraftResponseTestFactory.create();

		private final Bescheid bescheid = BescheidTestFactory.create();

		@DisplayName("on existing draft")
		@Nested
		class TestOnExistingDraft {

			@BeforeEach
			void mock() {
				when(service.findDraft(any())).thenReturn(Optional.of(bescheid));
				when(mapper.toGetBescheidDraftResponse(any())).thenReturn(response);
			}

			@Test
			void shouldCallService() {
				grpcService.getBescheidDraft(request, responseObserver);

				verify(service).findDraft(VorgangTestFactory.ID_STR);
			}

			@Test
			void shouldCallMapper() {
				grpcService.getBescheidDraft(request, responseObserver);

				verify(mapper).toGetBescheidDraftResponse(bescheid);
			}

			@Test
			void shouldCallNext() {
				grpcService.getBescheidDraft(request, responseObserver);

				verify(responseObserver).onNext(response);
			}

			@Test
			void shouldCallCompleted() {
				grpcService.getBescheidDraft(request, responseObserver);

				verify(responseObserver).onCompleted();
			}
		}

		@DisplayName("on missing draft")
		@Nested
		class TestOnMissingDraft {

			@Test
			void shouldCallNextWhenNoBescheid() {
				when(service.findDraft(any())).thenReturn(Optional.empty());

				grpcService.getBescheidDraft(request, responseObserver);

				verify(responseObserver).onNext(responseCaptor.capture());
				assertThat(responseCaptor.getValue().hasBescheid()).isFalse();
			}
		}
	}

	@Nested
	class TestGetConfig {

		private static final GrpcBescheidManagerConfigResponse RESPONSE = GrpcBescheidManagerConfigResponse.newBuilder().build();

		@Mock
		private GrpcBescheidManagerConfigRequest request;
		@Mock
		private StreamObserver<GrpcBescheidManagerConfigResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcBescheidManagerConfigResponse> responseCaptor;

		@Test
		void shouldCallBescheidService() {
			getConfig();

			verify(service).getConfig();
		}

		@Test
		void shouldCallGrpcBescheidMapper() {
			var config = BescheidManagerConfigTestFactory.create();
			when(service.getConfig()).thenReturn(config);

			getConfig();

			verify(mapper).fromBescheidManagerConfig(config);
		}

		@Test
		void shouldCallNext() {
			when(mapper.fromBescheidManagerConfig(any())).thenReturn(RESPONSE);

			getConfig();

			verify(responseObserver).onNext(RESPONSE);
		}

		@Test
		void shouldCallCompleted() {
			getConfig();

			verify(responseObserver).onCompleted();
		}

		private void getConfig() {
			grpcService.getConfig(request, responseObserver);
		}
	}

	@Nested
	class TestGetAll {

		private final GrpcGetAllBescheidResponse response = GrpcGetAllBescheidResponse.newBuilder().build();

		private final GrpcGetAllBescheidRequest request = GrpcGetAllBescheidRequestTestFactory.create();

		@Mock
		private StreamObserver<GrpcGetAllBescheidResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcGetAllBescheidResponse> responseCaptor;
		@Captor
		private ArgumentCaptor<VorgangId> vorgangIdCaptor;

		private final Stream<Bescheid> bescheide = Stream.of(BescheidTestFactory.create());

		@BeforeEach
		void init() {
			when(service.findAll(any())).thenReturn(bescheide);
			doReturn(response).when(grpcService).buildGetAllBescheidResponse(any());
		}

		@Test
		void shouldCallFindAllBescheid() {
			getAll();

			verify(service).findAll(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallBuildResponse() {
			getAll();

			verify(grpcService).buildGetAllBescheidResponse(bescheide);
		}

		@Test
		void shouldCallOnNext() {
			getAll();

			verify(responseObserver).onNext(response);
		}

		@Test
		void shouldCallOnCompleted() {
			getAll();

			verify(responseObserver).onCompleted();
		}

		private void getAll() {
			grpcService.getAll(request, responseObserver);
		}
	}

	@Nested
	class TestBuildResponseAllBescheid {

		private final Bescheid bescheid = BescheidTestFactory.create();
		private final Stream<Bescheid> bescheide = Stream.of(bescheid);

		private final GrpcBescheid grpcBescheid = GrpcBescheidTestFactory.create();

		@BeforeEach
		void init() {
			when(mapper.toBescheid(any())).thenReturn(grpcBescheid);
		}

		@Test
		void shouldCallMapper() {
			grpcService.buildGetAllBescheidResponse(bescheide);

			verify(mapper).toBescheid(bescheid);
		}

		@Test
		void shouldReturnResponse() {
			var response = grpcService.buildGetAllBescheidResponse(bescheide);

			assertThat(response.getBescheidList()).containsExactly(grpcBescheid);
		}
	}

	@DisplayName("Get bescheid")
	@Nested
	class TestGetBescheid {

		private final Bescheid bescheid = BescheidTestFactory.create();
		private final GrpcGetBescheidResponse response = GrpcGetBescheidResponseTestFactory.create();

		@Mock
		private StreamObserver<GrpcGetBescheidResponse> responseObserver;

		@BeforeEach
		void mock() {
			when(service.getBescheid(any())).thenReturn(bescheid);
			when(mapper.toGetBescheidResponse(any())).thenReturn(response);
		}

		@Test
		void shouldCallService() {
			getBescheid();

			verify(service).getBescheid(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallMapper() {
			getBescheid();

			verify(mapper).toGetBescheidResponse(bescheid);
		}

		@Test
		void shouldCallOnNext() {
			getBescheid();

			verify(responseObserver).onNext(response);
		}

		@Test
		void shouldCallCompleted() {
			getBescheid();

			verify(responseObserver).onCompleted();
		}

		private void getBescheid() {
			grpcService.getBescheid(GrpcGetBescheidRequestTestFactory.create(), responseObserver);
		}
	}
}
